﻿//
//  Conformance tests for NIRE.
//  2021/09/18; version 1.002
//
//  Macro Options:
//    -DUTF8: test with UTF-8 string.
//    -DUTF16: test with UTF-16 string.
//    -DUTF32: test with UTF-32 string.
//

#if !defined(UTF8) && !defined(UTF16) && !defined(UTF32)
#error Specify -DUTF8, -DUTF16, or -DUTF32.
#else

#include <cstdio>
#include <cstring>
#include <cwchar>
#include <ctime>
#include <string>
#include <vector>
#ifdef DUP_CHECK
#include <map>
#endif

#ifndef __cpp_unicode_characters
#define PRE_CPP11
#endif
#ifndef __cpp_char8_t
#define PRE_CPP20
#endif

#ifdef PRE_CPP11
	#if !defined(WCHAR_MAX) || (WCHAR_MAX < 0xFFFF) || (L'\u3000' != 0x3000)
	#error("Apparently this compiler cannot handle Unicode string literals.")
	#endif
typedef unsigned short char16_t;
typedef unsigned long char32_t;
#pragma message("char16_t and char32_t prepared for pre-C++11 compiler.")
#define NIRE_CPP11_CHAR1632_ENABLED
#endif

#ifdef PRE_CPP20
typedef unsigned char char8_t;
#pragma message("char8_t prepared for pre-C++20 compiler.")
#define NIRE_CPP20_CHAR8_ENABLED
#endif

#include "../nire.hpp"

#if defined(UTF16)
	#ifdef PRE_CPP11
		#if (WCHAR_MAX < 0x10FFFF)
		#define STRLIT(x) reinterpret_cast<const char16_t *>(L##x)
		#else
		#define STRLIT(x) utfconv::convert_utf32_to_utf16<wchar_t>(L##x)
		#endif
	#else
	#define STRLIT(x) u##x
	//#define CHRLIT(x) u##x
	#endif
typedef char16_t char_type;
#pragma message("char16_t used for UTF-16.")
#elif defined(UTF32)
	#ifdef PRE_CPP11
		#if (WCHAR_MAX < 0x10FFFF)
		#define STRLIT(x) utfconv::convert_utf16_to_utf32<wchar_t>(L##x)
		#else
		#define STRLIT(x) reinterpret_cast<const char32_t *>(L##x)
		#endif
	#else
	#define STRLIT(x) U##x
	//#define CHRLIT(x) U##x
	#endif
typedef char32_t char_type;
#pragma message("char32_t used for UTF-32.")
#else
//#elif defined(UTF8)
//	#ifndef UTF8
//	#define UTF8
//	#endif
	#ifdef PRE_CPP20
		#ifdef PRE_CPP11
		#define STRLIT(x) reinterpret_cast<const char8_t *>(utfconv::convert_to_utf8c(L##x).c_str())
		#else
		#define STRLIT(x) reinterpret_cast<const char8_t *>(utfconv::convert_utf32_to_utf8c<char32_t>(U##x).c_str())
		#endif
	#else
	#define STRLIT(x) u8##x
	//#define CHRLIT(x) u8##x
	#endif
typedef char8_t char_type;
#pragma message("char8_t used for UTF-8.")
#endif


namespace utfconv
{
template <typename charT>
int u32ctou8c(char out[4], const charT u32)
{
	//  UTF-32 -> UTF-8.
	if (u32 < 0x80)	//  00..7F
	{
		out[0] = static_cast<char>(u32);
		return 1;
	}
	else if (u32 < 0x800)	//  80..7FF
	{
		out[0] = static_cast<char>(((u32 >> 6) & 0x1f) | 0xc0);
		out[1] = static_cast<char>((u32 & 0x3f) | 0x80);
		return 2;
	}
	else if (u32 < 0x10000)	//  800..FFFF
	{
		out[0] = static_cast<char>(((u32 >> 12) & 0x0f) | 0xe0);
		out[1] = static_cast<char>(((u32 >> 6) & 0x3f) | 0x80);
		out[2] = static_cast<char>((u32 & 0x3f) | 0x80);
		return 3;
	}
//	else	//  10000..10FFFF
//	{
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4333)	//  Warns against the amount of right shift when charT is 16 bit.
#endif
		out[0] = static_cast<char>(((u32 >> 18) & 0x7) | 0xf0);
#ifdef _MSC_VER
#pragma warning(pop)
#endif
		out[1] = static_cast<char>(((u32 >> 12) & 0x3f) | 0x80);
		out[2] = static_cast<char>(((u32 >> 6) & 0x3f) | 0x80);
		out[3] = static_cast<char>((u32 & 0x3f) | 0x80);
		return 4;
//	}
}

template <typename Char16>
std::string convert_utf16_to_utf8c(const std::basic_string<Char16> &u16)
{
	std::string out;
	char utf8seq[4];

	for (typename std::basic_string<Char16>::size_type index = 0; index < u16.size(); ++index)
	{
		const Char16 ucp = u16[index];

		//  UTF-16 -> UTF-32.
		if ((ucp & 0xdc00) == 0xd800)
		{
			const Char16 surtail = u16[++index];

			if ((surtail & 0xdc00) == 0xdc00)
			{
				const Char16 surlead = ((ucp & 0x3ff) + 0x40);
				//  110110aa aabbbbcc  110111cc ddddeeee
				//  11110aaa 10aabbbb  10ccccdd 10ddeeee
				out.push_back(static_cast<char>(((surlead >> 8) & 7) | 0xf0));
				out.push_back(static_cast<char>(((surlead >> 2) & 0x3f) | 0x80));
				out.push_back(static_cast<char>(0x80 | ((surlead << 4) & 0x30) | ((surtail >> 6) & 0xf)));
				out.push_back(static_cast<char>(0x80 | (surtail & 0x3f)));
				continue;
			}
			else
				--index;
		}
		const int seqlen = u32ctou8c(utf8seq, ucp);
		out.append(utf8seq, seqlen);
	}
	return out;
}

template <typename Char32>
std::string convert_utf32_to_utf8c(const std::basic_string<Char32> &u32)
{
	std::string out;
	char utf8seq[4];

	for (typename std::basic_string<Char32>::size_type index = 0; index < u32.size(); ++index)
	{
		const int seqlen = u32ctou8c(utf8seq, u32[index]);
		out.append(utf8seq, seqlen);
	}

	return out;
}

template <typename Char16>
std::basic_string<char32_t> convert_utf16_to_utf32(const std::basic_string<Char16> &u16)
{
	std::basic_string<char32_t> out;

	for (typename std::basic_string<Char16>::size_type index = 0; index < u16.size(); ++index)
	{
		char32_t ucp = u16[index];

		//  UTF-16 -> UTF-32.
		if ((ucp & 0xdc00) == 0xd800)
		{
			if (++index < u16.size())
			{
				const char32_t surtail = u16[index];

				if ((surtail & 0xdc00) == 0xdc00)
					ucp = (((ucp & 0x3ff) << 10) | (surtail & 0x3ff)) + 0x10000;
				else
					--index;
			}
		}
		out += ucp;
	}
	return out;
}

template <typename Char32>
std::basic_string<char16_t> convert_utf32_to_utf16(const std::basic_string<Char32> &u32)
{
	std::basic_string<char16_t> out;

	for (typename std::basic_string<Char32>::size_type index = 0; index < u32.size(); ++index)
	{
		char32_t ucp = u32[index];

		//  UTF-32 -> UTF-16.
		if (ucp >= 0x10000)
		{
			const char32_t ls = (((ucp >> 10) - 0x40) & 0x3ff)| 0xd800;
			out.append(1, static_cast<char16_t>(ls));
			ucp = (ucp & 0x3ff) | 0xdc00;
		}
		out.append(1, static_cast<char16_t>(ucp));
	}
	return out;
}

std::string convert_to_utf8c(const std::string &s)
{
	return s;
}
std::string convert_to_utf8c(const std::wstring &ws)
{
#if (WCHAR_MAX < 0x10FFFF)
	return convert_utf16_to_utf8c(ws);
#else
	return convert_utf32_to_utf8c(ws);
#endif
}

std::string convert_to_utf8c(const std::basic_string<char8_t> &u8)
{
	return std::string(u8.begin(), u8.end());
}
std::string convert_to_utf8c(const std::basic_string<char16_t> &u16)
{
	return convert_utf16_to_utf8c(u16);
}
std::string convert_to_utf8c(const std::basic_string<char32_t> &u32)
{
	return convert_utf32_to_utf8c(u32);
}
}	//  namespace utfconv


template <typename charT>
bool conf_test(
	const bool type_regexp,	//  true: RegExp.exec(), false: String.match().
	std::size_t *const retval,
	const std::basic_string<charT> &str,
	const std::basic_string<charT> &exp,
	const std::vector<std::basic_string<charT> > &expected,
	const unsigned int /* max */,
	const nire::flag::type so = nire::flag::noflags,
	const std::size_t offset = 0)
{
	std::string flags;
	typedef std::basic_string<charT> string_type;
#ifdef REGEX_TYPE
	typedef nire::REGEX_TYPE regex_type;
#else
	typedef nire::regexp<charT> regex_type;
#endif
	regex_type re;
	typedef nire::re_match<charT> match_type;
	match_type mr;
	bool b = false;
	unsigned int num_of_failures = 0;
	const std::string strfc(utfconv::convert_to_utf8c(str));	//  For console.
	const std::string expfc(utfconv::convert_to_utf8c(exp));

	if (so & nire::flag::global)	//match_global)
		flags += 'g';

	if (so & nire::flag::icase)
		flags += 'i';

	if (so & nire::flag::multiline)
		flags += 'm';

	if (so & nire::flag::dotall)
		flags += 's';

	if (so & nire::flag::sticky)	//match_continuous)
		flags += 'y';

#ifdef DUP_CHECK
	static std::map<std::string, bool> tried;
	const std::string comb(flags + strfc + expfc);
	if (tried.count(comb))
		std::fprintf(stdout, "[Warning] /%s/%s.exec(\"%s\") has been checked.\n", expfc.c_str(), flags.c_str(), strfc.c_str());
	else
		tried[comb] = true;
#endif

	try
	{
		const bool lastindex_used = (so & (nire::flag::global | nire::flag::sticky)) != 0;
//		const charT *const begin = str.c_str() + ((so & nire::flag::global) ? 0 : offset);
		const charT *const begin = str.c_str() + (lastindex_used ? 0 : offset);
		const std::size_t len = static_cast<std::size_t>(str.size());
		string_type matched;
		string_type msg;

		re.assign(exp, so);
//		re.lastIndex = (so & nire::flag::global) ? offset : 0;
		re.lastIndex = lastindex_used ? offset : 0;

		const clock_t st = std::clock();

//		for (unsigned int i = 0; i < max; i++)
			b = re.exec(mr, begin, len);

		const clock_t ed = std::clock();

//		std::fprintf(stdout, "\t\"%s\" =~ /%s/%s;\n", strfc.c_str(), expfc.c_str(), flags.c_str());	//  Perl 5 style.
		if (type_regexp)
			std::fprintf(stdout, "\t/%s/%s.exec(\"%s\");\n", expfc.c_str(), flags.c_str(), strfc.c_str());	//  ECMAScript style.
		else
			std::fprintf(stdout, "\t\"%s\".match(/%s/%s);\n", strfc.c_str(), expfc.c_str(), flags.c_str());

		std::fprintf(stdout, "\t%s (%ld msec)\n", b ? "Found" : "Not Found", static_cast<long>(static_cast<double>(ed - st) * 1000 / CLOCKS_PER_SEC));

		if (!(so & nire::flag::global) || type_regexp)
		{
			for (typename match_type::size_type i = 0; i < mr.length(); ++i)
			{
				if (i)
					std::fprintf(stdout, "\t$%u = ", i);
				else
					std::fputs("\t$& = ", stdout);
				if (mr[i].matched())
				{
					matched = mr[i].str();
//					msg = CHRLIT('"') + matched + CHRLIT('"') + STRLIT(" (%u-%u)");
					msg = STRLIT("\"") + matched + STRLIT("\"") + STRLIT(" (%u-%u)");
				}
				else
					msg = matched = STRLIT("(undefined)");

				if (expected.size())
				{
					if (i < expected.size())
					{
						if (matched == expected.operator[](i))
							msg += STRLIT("; passed!");
						else
						{
							msg += STRLIT("; failed... (expected: \"") + expected.operator[](i) + STRLIT("\")");
							++num_of_failures;
						}
					}
					else
					{
						msg += STRLIT("; failed...");	//  Should not exist.
						++num_of_failures;
					}
				}
//				msg += CHRLIT('\n');
				msg += STRLIT("\n");
				const std::string msgfc(utfconv::convert_to_utf8c(msg));
				std::fprintf(stdout, msgfc.c_str(), mr.index(i), mr[i].length());
			}

			if (!num_of_failures && expected.size() != mr.length())
				++num_of_failures;

			std::fprintf(stdout, "Result: %s.\n\n", num_of_failures ? "failed" : "passed");
		}
		else	//  (so & nire::flag::global) && !type_regexp
		{
			unsigned int count = 0;

			for (;; ++count)
			{
				if (mr.length())
				{
					std::fprintf(stdout, "\t$&[%u] = ", count);

					if (mr[0].matched())
					{
						matched = mr[0].str();
//						msg = CHRLIT('"') + matched + CHRLIT('"') + STRLIT(" (%u-%u)");
						msg = STRLIT("\"") + matched + STRLIT("\"") + STRLIT(" (%u-%u)");
					}
					else
						msg = matched = STRLIT("(undefined)");

					if (expected.size())
					{
						if (count < expected.size())
						{
							if (matched == expected.operator[](count))
								msg += STRLIT("; passed!");
							else
							{
								msg += STRLIT("; failed... (expected: \"") + expected.operator[](count) + STRLIT("\")");
								++num_of_failures;
							}
						}
						else
						{
							msg += STRLIT("; failed...");	//  Should not exist.
							++num_of_failures;
						}
					}
//					msg += CHRLIT('\n');
					msg += STRLIT("\n");
					const std::string msgfc(utfconv::convert_to_utf8c(msg));
					std::fprintf(stdout, msgfc.c_str(), mr.index(0), mr[0].length());

//					for (unsigned int i = 0; i < max; i++)
						b = re.exec(mr, begin, len);
				}
				else
					break;
			}

			if (!num_of_failures && count != static_cast<unsigned int>(expected.size()))
				++num_of_failures;

			std::fprintf(stdout, "Result: %s.\n\n", num_of_failures ? "failed" : "passed");
		}
		if (retval)
			*retval = re.lastIndex;
		return num_of_failures == 0;
	}
	catch (const nire::re_error &e)
	{
		std::fprintf(stdout, "Error (re_error): %d \"%s\"\nResult: failed.\n\n", e.code(), e.what());
	}
	catch (const std::exception &e)
	{
		std::fprintf(stdout, "Error (std::exception): \"%s\"\nResult: failed.\n\n", e.what());
	}
	return false;
}


int main()
{
	typedef std::basic_string<char_type> string_type;
	const unsigned int count = 100000;
	string_type exp;
	string_type str;
	std::vector<string_type> expected;
	unsigned int num_of_tests = 0;
	unsigned int num_of_tests_passed = 0;
	unsigned int skipped = 0;
	const nire::flag::type syntax_default = nire::flag::ECMAScript;
	const nire::flag::type icase = nire::flag::icase;
	const nire::flag::type multiline = nire::flag::multiline;
	const nire::flag::type dotall = nire::flag::dotall;
	const nire::flag::type sticky = nire::flag::sticky;
	const nire::flag::type global = nire::flag::global;
	std::size_t retval;

#define TYPE_REGEXP_EXEC true
#define TYPE_STRING_MATCH false

	std::fputs("Test 1 (ECMAScript 2021 Language Specification 22.2.2.3, NOTE).\n", stdout);
	str = STRLIT("abc");
	exp = STRLIT("((a)|(ab))((c)|(bc))");
	expected.resize(7);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("a");
	expected[3] = STRLIT("(undefined)");
	expected[4] = STRLIT("bc");
	expected[5] = STRLIT("(undefined)");
	expected[6] = STRLIT("bc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 2a (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 2).\n", stdout);
	str = STRLIT("abcdefghi");
	exp = STRLIT("a[a-z]{2,4}");
	expected.resize(1);
	expected[0] = STRLIT("abcde");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 2b (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 2).\n", stdout);
	str = STRLIT("abcdefghi");
	exp = STRLIT("a[a-z]{2,4}?");
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 3 (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 2).\n", stdout);
	str = STRLIT("aabaac");
	exp = STRLIT("(aa|aabaac|ba|b|c)*");
	expected.resize(2);
	expected[0] = STRLIT("aaba");
	expected[1] = STRLIT("ba");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 4 (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 3).\n", stdout);
	str = STRLIT("zaacbbbcac");
	exp = STRLIT("(z)((a+)?(b+)?(c))*");
	expected.resize(6);
	expected[0] = STRLIT("zaacbbbcac");
	expected[1] = STRLIT("z");
	expected[2] = STRLIT("ac");
	expected[3] = STRLIT("a");
	expected[4] = STRLIT("(undefined)");
	expected[5] = STRLIT("c");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 5a (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 4).\n", stdout);
	str = STRLIT("b");
	exp = STRLIT("(a*)*");
	expected.resize(2);
	expected[0] = STRLIT("");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 5b (ECMAScript 2021 Language Specification 22.2.2.5.1, NOTE 4).\n", stdout);
	str = STRLIT("baaaac");
	exp = STRLIT("(a*)b\\1+");
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 6a (ECMAScript 2021 Language Specification 22.2.2.8.2, NOTE 2).\n", stdout);
	str = STRLIT("baaabac");
	exp = STRLIT("(?=(a+))");
	expected[0] = STRLIT("");
	expected[1] = STRLIT("aaa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 6b (ECMAScript 2021 Language Specification 22.2.2.8.2, NOTE 2).\n", stdout);
	str = STRLIT("baaabac");
	exp = STRLIT("(?=(a+))a*b\\1");
	expected[0] = STRLIT("aba");
	expected[1] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Test 7 (ECMAScript 2021 Language Specification 22.2.2.8.2, NOTE 3).\n", stdout);
	str = STRLIT("baaabaac");
	exp = STRLIT("(.*?)a(?!(a+)b\\2c)\\2(.*)");
	expected.resize(4);
	expected[0] = STRLIT("baaabaac");
	expected[1] = STRLIT("ba");
	expected[2] = STRLIT("(undefined)");
	expected[3] = STRLIT("abaac");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("TestB 01.\n", stdout);
	            //0123456
	str = STRLIT("aaaabaa");
	exp = STRLIT("^(.*)*b\\1$");
	expected.resize(2);
	expected[0] = STRLIT("aaaabaa");
	expected[1] = STRLIT("aa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 02.\n", stdout);
	            //0123456
	str = STRLIT("aaaabaaaa");
	exp = STRLIT("^(.*)*b\\1\\1$");
	expected[0] = STRLIT("aaaabaaaa");
	expected[1] = STRLIT("aa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 03.\n", stdout);
	            //01
	str = STRLIT("ab");
	exp = STRLIT("(.*?)*b\\1");
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 04.\n", stdout);
	            //01234567
	str = STRLIT("acaaabbb");
	exp = STRLIT("(a(.)a|\\2(.)b){2}");
	expected.resize(4);
	expected[0] = STRLIT("aaabb");
	expected[1] = STRLIT("bb");
	expected[2] = STRLIT("(undefined)");
	expected[3] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 05.\n", stdout);
	str = STRLIT("aabbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbaaaaaa");
	exp = STRLIT("(a*)(b)*\\1\\1\\1");
	expected.resize(3);
	expected[0] = STRLIT("aabbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbaaaaaa");
	expected[1] = STRLIT("aa");
	expected[2] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 06.\n", stdout);
	str = STRLIT("aaaaaaaaaab");
	exp = STRLIT("(.*)*b");
	expected.resize(2);
	expected[0] = STRLIT("aaaaaaaaaab");
	expected[1] = STRLIT("aaaaaaaaaa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 06a.\n", stdout);
	str = STRLIT("aaaaaaaaaab");
	exp = STRLIT("(.*)+b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 10))	//  The same results expected.
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 06b.\n", stdout);
	str = STRLIT("aaaaaaaaaab");
	exp = STRLIT("(.*){2,}b");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 07.\n", stdout);
	str = STRLIT("aaaaaaaaaabc");
	exp = STRLIT("(?=(a+))(abc)");
	expected.resize(3);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("TestB 08.\n", stdout);
	str = STRLIT("1234-5678-1234-456");
	exp = STRLIT("(\\d{4}[-]){3}\\d{3,4}");
	expected.resize(2);
	expected[0] = STRLIT("1234-5678-1234-456");
	expected[1] = STRLIT("1234-");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, count * 5))
		++num_of_tests_passed;
	++num_of_tests;

	//  icase.
	str = STRLIT("Σσςアイウ𛀀オやゆ𛀁よ");

	std::fputs("icase 01: sigma, case-sensitive.\n", stdout);
	exp = STRLIT("σ+");
	expected.resize(1);
	expected[0] = STRLIT("σ");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("icase 02: sigma, nocase.\n", stdout);
	exp = STRLIT("σ+");
	expected.resize(1);
	expected[0] = STRLIT("Σσς");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	//  Unicode property.
	std::fputs("Unicode property 01: kana.\n", stdout);
	exp = STRLIT("(?:(\\p{sc=Hiragana}+)|(\\p{sc=Katakana}+))+");
	expected.resize(3);
	expected[0] = STRLIT("アイウ𛀀オやゆ𛀁よ");
	expected[1] = STRLIT("やゆ𛀁よ");
	expected[2] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Unicode property 02: kana in charclass.\n", stdout);
	exp = STRLIT("[\\p{sc=Hiragana}\\p{sc=Katakana}]+");
	expected.resize(1);
	expected[0] = STRLIT("アイウ𛀀オやゆ𛀁よ");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Unicode property 03: kana in negated charclass.\n", stdout);
	exp = STRLIT("[^\\p{sc=Hiragana}]+");
	expected.resize(1);
	expected[0] = STRLIT("Σσςアイウ𛀀オ");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Broken UTF-8 & sticky flag 01a.\n", stdout);
	exp = STRLIT("[\\w ]+");
	str = STRLIT("ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789 abcdefghijklmnopqrstuvwxyz");
	expected.clear();
	expected.push_back(STRLIT("ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789 abcdefghijklmnopqrstuvwxyz"));
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag 02a: match with empty string.\n", stdout);
	exp = STRLIT("^$");
	str = STRLIT("");
	expected.resize(1);
	expected[0] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag 02b: an invalid UTF-8 code unit is not transparent.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("^$");
	str = STRLIT("\x80");
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag 02c: failure as above, even without sticky flag.\n", stdout);
#if defined(UTF8)
//	exp = STRLIT("^$");
//	str = STRLIT("\x80");
//	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag 03a: failure, even without ^.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("$");
	str = STRLIT("\x80");
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag 03b: match if without ^ and sticky flag.\n", stdout);
#if defined(UTF8)
//	exp = STRLIT("$");
//	str = STRLIT("\x80");
	expected.resize(1);
	expected[0] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif


	std::fputs("Broken UTF-8 & sticky flag & BMH 01a.\n", stdout);
	exp = STRLIT("ABC");
	str = STRLIT("ABC");
	expected.resize(1);
	expected[0] = STRLIT("ABC");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag & BMH 01b: icase.\n", stdout);
	exp = STRLIT("ABC");
	str = STRLIT("abc");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase | sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag & BMH 02a: failure because of invalid UTF-8.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("ABC");
//	str = STRLIT("\x80""ABC");
	str = STRLIT("ABC");
	str.insert(0, 1, 0x80);
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag & BMH 02b: same as above, icase.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("ABC$");
//	str = STRLIT("\x80""abc");
	str = STRLIT("abc");
	str.insert(0, 1, 0x80);
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase | sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag & BMH 02c: prevented by $.\n", stdout);
	exp = STRLIT("ABC$");
	str = STRLIT("ABCD");
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag & BMH 03a.\n", stdout);
	exp = STRLIT("ABC");
	str = STRLIT("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	expected.resize(1);
	expected[0] = STRLIT("ABC");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag & BMH 03b: icase.\n", stdout);
	exp = STRLIT("ABC");
	str = STRLIT("abcdefghijklmnopqrstuvwxyz");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase | sticky))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Broken UTF-8 & sticky flag & BMH 04a: failure because of invalid UTF-8.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("ABC");
//	str = STRLIT("\x80""ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	str = STRLIT("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	str.insert(0, 1, 0x80);
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif

	std::fputs("Broken UTF-8 & sticky flag & BMH 04b: same as above, icase.\n", stdout);
#if defined(UTF8)
	exp = STRLIT("ABC");
//	str = STRLIT("\x80""abcdefghijklmnopqrstuvwxyz");
	str = STRLIT("abcdefghijklmnopqrstuvwxyz");
	str.insert(0, 1, 0x80);
	expected.clear();
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase | sticky))
		++num_of_tests_passed;
	++num_of_tests;
#else
	std::fputs("[Caution] This test is available only in UTF-8. Skipped...\n\n", stdout);
	++skipped;
#endif


	//  Named group.
	std::fputs("Named capture 01.\n", stdout);
	exp = STRLIT("(?<year>\\d{4})/(?<month>\\d{2})/(?<日>\\d{2})");
	str = STRLIT("2000/December/31 2000-12-31 2000/12/31");
	expected.resize(4);
	expected[0] = STRLIT("2000/12/31");
	expected[1] = STRLIT("2000");
	expected[2] = STRLIT("12");
	expected[3] = STRLIT("31");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Named capture 02.\n", stdout);
	exp = STRLIT("(?<year>\\d{4})-(?<月>\\d{2})-(?<day>\\d{2})\\s+\\k<year>/\\k<月>/\\k<day>\\s+\\3\\2\\1");
	str = STRLIT("2000/December/31 2000-12-31 2000/12/31 31122000");
	expected.resize(4);
	expected[0] = STRLIT("2000-12-31 2000/12/31 31122000");
	expected[1] = STRLIT("2000");
	expected[2] = STRLIT("12");
	expected[3] = STRLIT("31");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	//  lastIndex.
	std::fputs("lastIndex 01: search starts at 10th element and fails (no global or sticky flag).\n", stdout);
	exp = STRLIT("(?<=^\\d+).+");
	str = STRLIT("0123456789abcdefghijklmnopqrstuvwxyz");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, syntax_default, 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("lastIndex 02: with global flag.\n", stdout);
	expected.resize(1);
	expected[0] = STRLIT("abcdefghijklmnopqrstuvwxyz");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, global, 10))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("lastIndex 03: with sticky flag.\n", stdout);
	exp = STRLIT("(?<=^\\d+).+");
	str = STRLIT("0123456789abcdefghijklmnopqrstuvwxyz");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, sticky, 10))
		++num_of_tests_passed;
	++num_of_tests;

	//  https://github.com/tc39/proposal-regexp-lookbehind
	std::fputs("Lookbehind 01.\n", stdout);
	exp = STRLIT("(?<=(\\d+)(\\d+))$");
	str = STRLIT("1053");
	expected.resize(3);
	expected[0] = STRLIT("");
	expected[1] = STRLIT("1");
	expected[2] = STRLIT("053");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	//  https://github.com/tc39/test262/tree/master/test/built-ins/RegExp/lookBehind
	std::fputs("Lookbehind 02a: alternations.js #1.\n", stdout);
	exp = STRLIT(".*(?<=(..|...|....))(.*)");
	str = STRLIT("xabcd");
	expected.resize(3);
	expected[0] = STRLIT("xabcd");
	expected[1] = STRLIT("cd");
	expected[2] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 02b: alternations.js #2.\n", stdout);
	exp = STRLIT(".*(?<=(xx|...|....))(.*)");
	str = STRLIT("xabcd");
	expected.resize(3);
	expected[0] = STRLIT("xabcd");
	expected[1] = STRLIT("bcd");
	expected[2] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 02c: alternations.js #3.\n", stdout);
	exp = STRLIT(".*(?<=(xx|...))(.*)");
	str = STRLIT("xxabcd");
	expected.resize(3);
	expected[0] = STRLIT("xxabcd");
	expected[1] = STRLIT("bcd");
	expected[2] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 02d: alternations.js #4.\n", stdout);
	exp = STRLIT(".*(?<=(xx|xxx))(.*)");
	str = STRLIT("xxabcd");
	expected.resize(3);
	expected[0] = STRLIT("xxabcd");
	expected[1] = STRLIT("xx");
	expected[2] = STRLIT("abcd");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 03a: back-references-to-captures.js #1.\n", stdout);
	exp = STRLIT("(?<=\\1(\\w))d");
	str = STRLIT("abcCd");
	expected.resize(2);
	expected[0] = STRLIT("d");
	expected[1] = STRLIT("C");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 03b: back-references-to-captures.js #2.\n", stdout);
	exp = STRLIT("(?<=\\1([abx]))d");
	str = STRLIT("abxxd");
	expected.resize(2);
	expected[0] = STRLIT("d");
	expected[1] = STRLIT("x");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 03c: back-references-to-captures.js #3.\n", stdout);
	exp = STRLIT("(?<=\\1(\\w+))c");
	str = STRLIT("ababc");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("ab");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 03d: back-references-to-captures.js #4.\n", stdout);
	exp = STRLIT("(?<=\\1(\\w+))c");
	str = STRLIT("ababbc");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 03e: back-references-to-captures.js #5.\n", stdout);
	exp = STRLIT("(?<=\\1(\\w+))c");
	str = STRLIT("ababdc");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 03f: back-references-to-captures.js #6.\n", stdout);
	exp = STRLIT("(?<=(\\w+)\\1)c");
	str = STRLIT("ababc");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("abab");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 04a: back-references.js #1.\n", stdout);
	exp = STRLIT("(.)(?<=(\\1\\1))");
	str = STRLIT("abb");
	expected.resize(3);
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("b");
	expected[2] = STRLIT("bb");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04b: back-references.js #2.\n", stdout);
	exp = STRLIT("(.)(?<=(\\1\\1))");
	str = STRLIT("abB");
	expected.resize(3);
	expected[0] = STRLIT("B");
	expected[1] = STRLIT("B");
	expected[2] = STRLIT("bB");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04c: back-references.js #3.\n", stdout);
	exp = STRLIT("((\\w)\\w)(?<=\\1\\2\\1)");
	str = STRLIT("aabAaBa");
	expected.resize(3);
	expected[0] = STRLIT("aB");
	expected[1] = STRLIT("aB");
	expected[2] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04d: back-references.js #4.\n", stdout);
	exp = STRLIT("(\\w(\\w))(?<=\\1\\2\\1)");
	str = STRLIT("aabAaBa");
	expected.resize(3);
	expected[0] = STRLIT("Ba");
	expected[1] = STRLIT("Ba");
	expected[2] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04e: back-references.js #5.\n", stdout);
	exp = STRLIT("(?=(\\w))(?<=(\\1)).");
	str = STRLIT("abaBbAa");
	expected.resize(3);
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("b");
	expected[2] = STRLIT("B");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04f: back-references.js #6.\n", stdout);
	exp = STRLIT("(?<=(.))(\\w+)(?=\\1)");
	str = STRLIT("  'foo'  ");
	expected.resize(3);
	expected[0] = STRLIT("foo");
	expected[1] = STRLIT("'");
	expected[2] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04g: back-references.js #7.\n", stdout);
	exp = STRLIT("(?<=(.))(\\w+)(?=\\1)");
	str = STRLIT("  \"foo\"  ");
	expected.resize(3);
	expected[0] = STRLIT("foo");
	expected[1] = STRLIT("\"");
	expected[2] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04h: back-references.js #8.\n", stdout);
	exp = STRLIT("(.)(?<=\\1\\1\\1)");
	str = STRLIT("abbb");
	expected.resize(2);
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04i: back-references.js #9.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("fababab");
	expected.resize(2);
	expected[0] = STRLIT("ab");
	expected[1] = STRLIT("ab");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04j: back-references.js #10.\n", stdout);
	exp = STRLIT("(?<=(.))(\\w+)(?=\\1)");
	str = STRLIT("  .foo\"  ");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04k: back-references.js #11.\n", stdout);
	exp = STRLIT("(.)(?<=\\1\\1\\1)");
	str = STRLIT("ab");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04l: back-references.js #12.\n", stdout);
	exp = STRLIT("(.)(?<=\\1\\1\\1)");
	str = STRLIT("abb");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04l: back-references.js #13.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("ab");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04m: back-references.js #14.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("abb");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04n: back-references.js #15.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("aabb");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04o: back-references.js #16.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("abab");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04p: back-references.js #17.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("fabxbab");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 04q: back-references.js #18.\n", stdout);
	exp = STRLIT("(..)(?<=\\1\\1\\1)");
	str = STRLIT("faxabab");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 05: captures-negative.js.\n", stdout);
	exp = STRLIT("(?<!(^|[ab]))\\w{2}");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("de");
	expected[1] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	str = STRLIT("abcdef");
	std::fputs("Lookbehind 06a: captures.js #1.\n", stdout);
	exp = STRLIT("(?<=(c))def");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("c");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06b: captures.js #2.\n", stdout);
	exp = STRLIT("(?<=(\\w{2}))def");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("bc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06c: captures.js #3.\n", stdout);
	exp = STRLIT("(?<=(\\w(\\w)))def");
	expected.resize(3);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("bc");
	expected[2] = STRLIT("c");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06d: captures.js #4.\n", stdout);
	exp = STRLIT("(?<=(\\w){3})def");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06e: captures.js #5.\n", stdout);
	exp = STRLIT("(?<=(bc)|(cd)).");
	expected.resize(3);
	expected[0] = STRLIT("d");
	expected[1] = STRLIT("bc");
	expected[2] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06f: captures.js #6.\n", stdout);
	exp = STRLIT("(?<=([ab]{1,2})\\D|(abc))\\w");
	expected.resize(3);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06g: captures.js #7.\n", stdout);
	exp = STRLIT("\\D(?<=([ab]+))(\\w)");
	expected.resize(3);
	expected[0] = STRLIT("ab");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06h: captures.js #8.\n", stdout);
	exp = STRLIT("(?<=b|c)\\w");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("d");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, global))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 06i: captures.js #9.\n", stdout);
	exp = STRLIT("(?<=[b-e])\\w{2}");
	expected.resize(2);
	expected[0] = STRLIT("cd");
	expected[1] = STRLIT("ef");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, global))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 07: do-not-backtrack.js.\n", stdout);
	exp = STRLIT("(?<=([abc]+)).\\1");
	str = STRLIT("abcdbc");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 08a: greedy-loop.js #1.\n", stdout);
	exp = STRLIT("(?<=(b+))c");
	str = STRLIT("abbbbbbc");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("bbbbbb");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 08b: greedy-loop.js #2.\n", stdout);
	exp = STRLIT("(?<=(b\\d+))c");
	str = STRLIT("ab1234c");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("b1234");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 08c: greedy-loop.js #3.\n", stdout);
	exp = STRLIT("(?<=((?:b\\d{2})+))c");
	str = STRLIT("ab12b23b34c");
	expected.resize(2);
	expected[0] = STRLIT("c");
	expected[1] = STRLIT("b12b23b34");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 09a: misc.js #1.\n", stdout);
	exp = STRLIT("(?<=$abc)def");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09b: misc.js #2.\n", stdout);
	exp = STRLIT("^f.o(?<=foo)$");
	str = STRLIT("fno");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09c: misc.js #3.\n", stdout);
	exp = STRLIT("^foo(?<!foo)$");
	str = STRLIT("foo");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09d: misc.js #4.\n", stdout);
	exp = STRLIT("^f.o(?<!foo)$");
	str = STRLIT("foo");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09e: misc.js #5.\n", stdout);
	exp = STRLIT("^foo(?<=foo)$");
	str = STRLIT("foo");
	expected.resize(1);
	expected[0] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09f: misc.js #6.\n", stdout);
	exp = STRLIT("^f.o(?<=foo)$");
	str = STRLIT("foo");
	expected.resize(1);
	expected[0] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09g: misc.js #7.\n", stdout);
	exp = STRLIT("^f.o(?<!foo)$");
	str = STRLIT("fno");
	expected.resize(1);
	expected[0] = STRLIT("fno");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09h: misc.js #8.\n", stdout);
	exp = STRLIT("^foooo(?<=fo+)$");
	str = STRLIT("foooo");
	expected.resize(1);
	expected[0] = STRLIT("foooo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09i: misc.js #9.\n", stdout);
	exp = STRLIT("^foooo(?<=fo*)$");
	str = STRLIT("foooo");
	expected.resize(1);
	expected[0] = STRLIT("foooo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09j: misc.js #10.\n", stdout);
	exp = STRLIT("(abc\\1)");
	str = STRLIT("abc");
	expected.resize(2);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09k: misc.js #11.\n", stdout);
	exp = STRLIT("(abc\\1)");
	str = STRLIT("abc\u1234");
	expected.resize(2);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09l: misc.js #12.\n", stdout);
	exp = STRLIT("(abc\\1)");
	str = STRLIT("abc");
	expected.resize(2);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 09m: misc.js #13.\n", stdout);
	exp = STRLIT("(abc\\1)");
	str = STRLIT("abc\u1234");
	expected.resize(2);
	expected[0] = STRLIT("abc");
	expected[1] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 10a: mutual-recursive.js #1.\n", stdout);
	exp = STRLIT("(?<=a(.\\2)b(\\1)).{4}");
	str = STRLIT("aabcacbc");
	expected.resize(3);
	expected[0] = STRLIT("cacb");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 10b: mutual-recursive.js #2.\n", stdout);
	exp = STRLIT("(?<=a(\\2)b(..\\1))b");
	str = STRLIT("aacbacb");
	expected.resize(3);
	expected[0] = STRLIT("b");
	expected[1] = STRLIT("ac");
	expected[2] = STRLIT("ac");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 10c: mutual-recursive.js #3.\n", stdout);
	exp = STRLIT("(?<=(?:\\1b)(aa)).");
	str = STRLIT("aabaax");
	expected.resize(2);
	expected[0] = STRLIT("x");
	expected[1] = STRLIT("aa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 10d: mutual-recursive.js #4.\n", stdout);
	exp = STRLIT("(?<=(?:\\1|b)(aa)).");
	str = STRLIT("aaaax");
	expected.resize(2);
	expected[0] = STRLIT("x");
	expected[1] = STRLIT("aa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	str = STRLIT("abcdef");
	std::fputs("Lookbehind 11a: negative.js #1.\n", stdout);
	exp = STRLIT("(?<!abc)\\w\\w\\w");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11b: negative.js #2.\n", stdout);
	exp = STRLIT("(?<!a.c)\\w\\w\\w");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11c: negative.js #3.\n", stdout);
	exp = STRLIT("(?<!a\\wc)\\w\\w\\w");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11d: negative.js #4.\n", stdout);
	exp = STRLIT("(?<!a[a-z])\\w\\w\\w");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11e: negative.js #5.\n", stdout);
	exp = STRLIT("(?<!a[a-z]{2})\\w\\w\\w");
	expected.resize(1);
	expected[0] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11f: negative.js #6.\n", stdout);
	exp = STRLIT("(?<!abc)def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11g: negative.js #7.\n", stdout);
	exp = STRLIT("(?<!a.c)def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11h: negative.js #8.\n", stdout);
	exp = STRLIT("(?<!a\\wc)def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11i: negative.js #9.\n", stdout);
	exp = STRLIT("(?<!a[a-z][a-z])def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11j: negative.js #10.\n", stdout);
	exp = STRLIT("(?<!a[a-z]{2})def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11k: negative.js #11.\n", stdout);
	exp = STRLIT("(?<!a{1}b{1})cde");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 11l: negative.js #12.\n", stdout);
	exp = STRLIT("(?<!a{1}[a-z]{2})def");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 12a: nested-lookaround.js #1.\n", stdout);
	exp = STRLIT("(?<=ab(?=c)\\wd)\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("ef");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 12b: nested-lookaround.js #2.\n", stdout);
	exp = STRLIT("(?<=a(?=([^a]{2})d)\\w{3})\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("ef");
	expected[1] = STRLIT("bc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 12c: nested-lookaround.js #3.\n", stdout);
	exp = STRLIT("(?<=a(?=([bc]{2}(?<!a{2}))d)\\w{3})\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("ef");
	expected[1] = STRLIT("bc");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 12d: nested-lookaround.js #4.\n", stdout);
	exp = STRLIT("^faaao?(?<=^f[oa]+(?=o))");
	str = STRLIT("faaao");
	expected.resize(1);
	expected[0] = STRLIT("faaa");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 12e: nested-lookaround.js #5.\n", stdout);
	exp = STRLIT("(?<=a(?=([bc]{2}(?<!a*))d)\\w{3})\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 13a: simple-fixed-length.js #1.\n", stdout);
	exp = STRLIT("^.(?<=a)");
	str = STRLIT("b");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13b: simple-fixed-length.js #2.\n", stdout);
	exp = STRLIT("^f\\w\\w(?<=\\woo)");
	str = STRLIT("boo");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13c: simple-fixed-length.js #3.\n", stdout);
	exp = STRLIT("^f\\w\\w(?<=\\woo)");
	str = STRLIT("fao");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13d: simple-fixed-length.js #4.\n", stdout);
	exp = STRLIT("^f\\w\\w(?<=\\woo)");
	str = STRLIT("foa");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13e: simple-fixed-length.js #5.\n", stdout);
	exp = STRLIT("^.(?<=a)");
	str = STRLIT("a");
	expected.resize(1);
	expected[0] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13f: simple-fixed-length.js #6.\n", stdout);
	exp = STRLIT("^f..(?<=.oo)");
	str = STRLIT("foo1");
	expected.resize(1);
	expected[0] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13g: simple-fixed-length.js #7.\n", stdout);
	exp = STRLIT("^f\\w\\w(?<=\\woo)");
	str = STRLIT("foo2");
	expected.resize(1);
	expected[0] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13h: simple-fixed-length.js #8.\n", stdout);
	exp = STRLIT("(?<=abc)\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13i: simple-fixed-length.js #9.\n", stdout);
	exp = STRLIT("(?<=a.c)\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13j: simple-fixed-length.js #10.\n", stdout);
	exp = STRLIT("(?<=a\\wc)\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13k: simple-fixed-length.js #11.\n", stdout);
	exp = STRLIT("(?<=a[a-z])\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("cde");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13l: simple-fixed-length.js #12.\n", stdout);
	exp = STRLIT("(?<=a[a-z][a-z])\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13m: simple-fixed-length.js #13.\n", stdout);
	exp = STRLIT("(?<=a[a-z]{2})\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13n: simple-fixed-length.js #14.\n", stdout);
	exp = STRLIT("(?<=a{1})\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("bcd");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13o: simple-fixed-length.js #15.\n", stdout);
	exp = STRLIT("(?<=a{1}b{1})\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("cde");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 13p: simple-fixed-length.js #16.\n", stdout);
	exp = STRLIT("(?<=a{1}[a-z]{2})\\w\\w\\w");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	str = STRLIT("abcdefghijklmn");
	std::fputs("Lookbehind 14a: sliced-strings.js #1.\n", stdout);
	exp = STRLIT("(?=(abcdefghijklmn))(?<=\\1)a");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 14b: sliced-strings.js #2.\n", stdout);
	exp = STRLIT("(?=(abcdefghijklmn))(?<=\\1)a");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 14c: sliced-strings.js #3.\n", stdout);
	exp = STRLIT("(?=(abcdefg))(?<=\\1)");
	str = STRLIT("bcdefgabcdefg");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 15a: start-of-line.js #1.\n", stdout);
	exp = STRLIT("(?<=^[^a-c]{3})def");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15b: start-of-line.js #2.\n", stdout);
	exp = STRLIT("\"^foooo(?<=^o+)$");
	str = STRLIT("foooo");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15c: start-of-line.js #3.\n", stdout);
	exp = STRLIT("\"^foooo(?<=^o*)$");
	str = STRLIT("foooo");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15d: start-of-line.js #4.\n", stdout);
	exp = STRLIT("(?<=^abc)def");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15e: start-of-line.js #5.\n", stdout);
	exp = STRLIT("(?<=^[a-c]{3})def");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15f: start-of-line.js #6.\n", stdout);
	exp = STRLIT("(?<=^[a-c]{3})def");
	str = STRLIT("xyz\nabcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, multiline))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15g: start-of-line.js #7.\n", stdout);
	exp = STRLIT("(?<=^)\\w+");
	str = STRLIT("ab\ncd\nefg");
	expected.resize(3);
	expected[0] = STRLIT("ab");
	expected[1] = STRLIT("cd");
	expected[2] = STRLIT("efg");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, multiline | global))
	{
		++num_of_tests_passed;
	}
	++num_of_tests;

	std::fputs("Lookbehind 15h: start-of-line.js #8.\n", stdout);
	exp = STRLIT("\\w+(?<=$)");
	str = STRLIT("ab\ncd\nefg");
	expected.resize(3);
	expected[0] = STRLIT("ab");
	expected[1] = STRLIT("cd");
	expected[2] = STRLIT("efg");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, multiline | global))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15i: start-of-line.js #9.\n", stdout);
	exp = STRLIT("(?<=^)\\w+(?<=$)");
	str = STRLIT("ab\ncd\nefg");
	expected.resize(3);
	expected[0] = STRLIT("ab");
	expected[1] = STRLIT("cd");
	expected[2] = STRLIT("efg");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, multiline | global))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15j: start-of-line.js #10.\n", stdout);
	exp = STRLIT("^foo(?<=^fo+)$");
	str = STRLIT("foo");
	expected.resize(1);
	expected[0] = STRLIT("foo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15k: start-of-line.js #11.\n", stdout);
	exp = STRLIT("^foooo(?<=^fo*)");
	str = STRLIT("foooo");
	expected.resize(1);
	expected[0] = STRLIT("foooo");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15l: start-of-line.js #12.\n", stdout);
	exp = STRLIT("^(f)oo(?<=^\\1o+)$");
	str = STRLIT("foo");
	expected.resize(2);
	expected[0] = STRLIT("foo");
	expected[1] = STRLIT("f");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15m: start-of-line.js #13.\n", stdout);
	exp = STRLIT("^(f)oo(?<=^\\1o+)$");
	str = STRLIT("foo");
	expected.resize(2);
	expected[0] = STRLIT("foo");
	expected[1] = STRLIT("f");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15n: start-of-line.js #14.\n", stdout);
	exp = STRLIT("^(f)oo(?<=^\\1o+).$");
	str = STRLIT("foo\u1234");
	expected.resize(2);
	expected[0] = STRLIT("foo\u1234");
	expected[1] = STRLIT("f");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, icase))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15o: start-of-line.js #15.\n", stdout);
	exp = STRLIT("(?<=^\\w+)def");
	str = STRLIT("abcdefdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 15p: start-of-line.js #16.\n", stdout);
	exp = STRLIT("(?<=^\\w+)def");
	str = STRLIT("abcdefdef");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("def");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, global))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 16a: sticky.js #1.\n", stdout);
	exp = STRLIT("(?<=^(\\w+))def");
	str = STRLIT("abcdefdef");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("abc");
	if (conf_test(TYPE_REGEXP_EXEC, &retval, str, exp, expected, 1, global))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 16b: sticky.js #2.\n", stdout);
	exp = STRLIT("(?<=^(\\w+))def");
	str = STRLIT("abcdefdef");
	expected.resize(2);
	expected[0] = STRLIT("def");
	expected[1] = STRLIT("abcdef");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, global, retval))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 16c: sticky.js #3.\n", stdout);
	exp = STRLIT("\\Bdef");
	str = STRLIT("abcdefdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, &retval, str, exp, expected, 1, global))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 16d: sticky.js #4.\n", stdout);
	str = STRLIT("abcdefdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_STRING_MATCH, NULL, str, exp, expected, 1, global, retval))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 17a: variable-length.js #1.\n", stdout);
	exp = STRLIT("(?<=[a|b|c]*)[^a|b|c]{3}");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 17b: variable-length.js #2.\n", stdout);
	exp = STRLIT("(?<=\\w*)[^a|b|c]{3}");
	str = STRLIT("abcdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("Lookbehind 18a: word-boundary.js #1.\n", stdout);
	exp = STRLIT("(?<=\\b)[d-f]{3}");
	str = STRLIT("abc def");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 18b: word-boundary.js #2.\n", stdout);
	exp = STRLIT("(?<=\\B)\\w{3}");
	str = STRLIT("ab cdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 18c: word-boundary.js #3.\n", stdout);
	exp = STRLIT("(?<=\\B)(?<=c(?<=\\w))\\w{3}");
	str = STRLIT("ab cdef");
	expected.resize(1);
	expected[0] = STRLIT("def");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("Lookbehind 18d: word-boundary.js #4.\n", stdout);
	exp = STRLIT("(?<=\\b)[d-f]{3}");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	//  https://github.com/tc39/test262/tree/master/test/built-ins/RegExp/dotall
	std::fputs("dotall flag 01: with-dotall-unicode.js.\n", stdout);
	exp = STRLIT(".*");
#ifndef _MSC_VER
	str = STRLIT("a3π\u2027\u0085\v\f\u180E\U00010300\n\r\u2028\u2029");
	expected.resize(1);
	expected[0] = str;
#else
	str = STRLIT("a3π\u2027\v\f\u180E\U00010300\n\r\u2028\u2029");
	expected.resize(1);
	expected[0] = str;
#endif
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1, dotall))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("dotall flag 02: without-dotall-unicode.js.\n", stdout);
	exp = STRLIT("(^.*)([\\s]+$)");
#ifndef _MSC_VER
	str = STRLIT("a3π\u2027\u0085\v\f\u180E\U00010300\n\r\u2028\u2029");	//  \uD800\uDFFF");
	expected.resize(3);
	expected[0] = str;
	expected[1] = STRLIT("a3π\u2027\u0085\v\f\u180E\U00010300");
#else
	str = STRLIT("a3π\u2027\v\f\u180E\U00010300\n\r\u2028\u2029");	//  \uD800\uDFFF");
	expected.resize(3);
	expected[0] = str;
	expected[1] = STRLIT("a3π\u2027\v\f\u180E\U00010300");
#endif
	expected[2] = STRLIT("\n\r\u2028\u2029");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	//  https://github.com/tc39/test262/tree/master/test/built-ins/RegExp/named-groups
	std::fputs("named-groups 01a: lookbehind.js #1.\n", stdout);
	exp = STRLIT("(?<=(?<a>\\w){3})f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("c");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

//	std::fputs("named-groups 01b: lookbehind.js #2.\n", stdout);

	std::fputs("named-groups 01c: lookbehind.js #2.\n", stdout);
	exp = STRLIT("(?<=(?<a>\\w){4})f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01d: lookbehind.js #3.\n", stdout);
	exp = STRLIT("(?<=(?<a>\\w)+)f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01e: lookbehind.js #4.\n", stdout);
	exp = STRLIT("(?<=(?<a>\\w){6})f");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01f: lookbehind.js #5.\n", stdout);
	exp = STRLIT("((?<=\\w{3}))f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01g: lookbehind.js #6.\n", stdout);
	exp = STRLIT("(?<a>(?<=\\w{3}))f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01h: lookbehind.js #7.\n", stdout);
	exp = STRLIT("(?<!(?<a>\\d){3})f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01i: lookbehind.js #8.\n", stdout);
	exp = STRLIT("(?<!(?<a>\\D){3})f");
	str = STRLIT("abcdef");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01j: lookbehind.js #9.\n", stdout);
	exp = STRLIT("(?<!(?<a>\\D){3})f|f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 01k: lookbehind.js #10.\n", stdout);
	exp = STRLIT("(?<a>(?<!\\D{3}))f|f");
	str = STRLIT("abcdef");
	expected.resize(2);
	expected[0] = STRLIT("f");
	expected[1] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02a: unicode-match.js #1.\n", stdout);
	exp = STRLIT("(?<a>.)(?<b>.)(?<c>.)\\k<c>\\k<b>\\k<a>");
	str = STRLIT("abccba");
	expected.resize(4);
	expected[0] = STRLIT("abccba");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("b");
	expected[3] = STRLIT("c");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02b: unicode-match.js #2.\n", stdout);
	exp = STRLIT("(?<b>b).\\1");
	str = STRLIT("bab");
	expected.resize(2);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02c: unicode-match.js #3.\n", stdout);
	exp = STRLIT("(.)(?<a>a)\\1\\2");
	str = STRLIT("baba");
	expected.resize(3);
	expected[0] = STRLIT("baba");
	expected[1] = STRLIT("b");
	expected[2] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02d: unicode-match.js #4.\n", stdout);
	exp = STRLIT("(.)(?<a>a)(?<b>\\1)(\\2)");
	str = STRLIT("baba");
	expected.resize(5);
	expected[0] = STRLIT("baba");
	expected[1] = STRLIT("b");
	expected[2] = STRLIT("a");
	expected[3] = STRLIT("b");
	expected[4] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02e: unicode-match.js #5.\n", stdout);
	exp = STRLIT("(?<lt><)a");
	str = STRLIT("<a");
	expected.resize(2);
	expected[0] = STRLIT("<a");
	expected[1] = STRLIT("<");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02f: unicode-match.js #6.\n", stdout);
	exp = STRLIT("(?<gt>>)a");
	str = STRLIT(">a");
	expected.resize(2);
	expected[0] = STRLIT(">a");
	expected[1] = STRLIT(">");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 02g: unicode-match.js #7.\n", stdout);
	exp = STRLIT("(?<a>.(?<b>.(?<c>.)))");
	str = STRLIT("bab");
	expected.resize(4);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("bab");
	expected[2] = STRLIT("ab");
	expected[3] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fputs("named-groups 03a: unicode-references.js #1.\n", stdout);
	exp = STRLIT("(?<b>.).\\k<b>");
	str = STRLIT("bab");
	expected.resize(2);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03b: unicode-references.js #2.\n", stdout);
	exp = STRLIT("(?<b>.).\\k<b>");
	str = STRLIT("baa");
	expected.resize(0);
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03c: unicode-references.js #3.\n", stdout);
	exp = STRLIT("(?<a>\\k<a>\\w)..");
	str = STRLIT("bab");
	expected.resize(2);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03d: unicode-references.js #4.\n", stdout);
	exp = STRLIT("\\k<a>(?<a>b)\\w\\k<a>");
	str = STRLIT("bab");
	expected.resize(2);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03e: unicode-references.js #5.\n", stdout);
	exp = STRLIT("(?<b>b)\\k<a>(?<a>a)\\k<b>");
	str = STRLIT("bab");
	expected.resize(3);
	expected[0] = STRLIT("bab");
	expected[1] = STRLIT("b");
	expected[2] = STRLIT("a");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03f: unicode-references.js #6.\n", stdout);
	exp = STRLIT("(?<a>a)(?<b>b)\\k<a>");
	str = STRLIT("aba");
	expected.resize(3);
	expected[0] = STRLIT("aba");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("b");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;

	std::fputs("named-groups 03g: unicode-references.js #7.\n", stdout);
	exp = STRLIT("(?<a>a)(?<b>b)\\k<a>|(?<c>c)");
	str = STRLIT("aba");
	expected.resize(4);
	expected[0] = STRLIT("aba");
	expected[1] = STRLIT("a");
	expected[2] = STRLIT("b");
	expected[3] = STRLIT("(undefined)");
	if (conf_test(TYPE_REGEXP_EXEC, NULL, str, exp, expected, 1))
		++num_of_tests_passed;
	++num_of_tests;


	std::fprintf(stdout, "Results of tests: %u/%u (%.1lf%%) passed. (%u skipped).\n", num_of_tests_passed, num_of_tests, static_cast<double>(num_of_tests_passed) * 100.0 / num_of_tests, skipped);

	return 0;
}

#endif
